#!/usr/bin/bash

# vim:ft=sh:fdm=marker:fmr={,}
# shellcheck disable=2034,2154

# archlabs installer library script file
# this file is not meant to be run directly
# sourcing this file in a non bash shell is not advised

# command used to install each bootloader
declare -Ag BCMDS=(
[syslinux]="syslinux-install_update -iam"
[grub]="grub-install --recheck --force"
[systemd-boot]="bootctl --path=/boot install"
)

# boot partition mount points for each bootloader
declare -Ag BMNTS=(
[BIOS-grub]="/boot"
[UEFI-grub]="/boot/efi"
[BIOS-syslinux]="/boot"
[UEFI-systemd-boot]="/boot"
)

# bootloader options with respective boot partition mountpoint
declare -Ag BOOTLDRS=(
[BIOS]="grub ${BMNTS[BIOS-grub]} syslinux ${BMNTS[BIOS-syslinux]}"
[UEFI]="systemd-boot ${BMNTS[UEFI-systemd-boot]} grub ${BMNTS[UEFI-grub]}"
)

setup_boot()
{
    tput civis
    if ! BOOTLDR="$(menubox "$_PrepMount" "$_MntBootBody" 0 0 0 ${BOOTLDRS[$SYS]})"; then
        return 1
    fi

    if [[ $BOOT_PART != "" ]]; then
        mount_boot_part || return 1
        setup_boot_device
    fi

    setup_${BOOTLDR} || return 1
}

setup_grub()
{
    EDIT_FILES[9]="/etc/default/grub"

    if [[ $SYS == 'BIOS' ]]; then
        if [[ $BOOT_DEVICE == "" ]]; then
            select_device 'boot' || return 1
        fi
        BCMDS[grub]+=" --target=i386-pc $BOOT_DEVICE && grub-mkconfig -o /boot/grub/grub.cfg"
    else
        if [[ $ROOT_PART =~ /dev/mapper && ! $LVM && ! $LUKS_PASS ]]; then
            luks_pass "$_LuksOpen" "$ROOT_PART" || return 1
        fi

        # the mount mess is needed for os-prober to work properly in the chroot
        BCMDS[grub]="mkdir -p /run/udev && mkdir -p /run/lvm &&
              mount --bind /hostrun/udev /run/udev && mount --bind /hostrun/lvm /run/lvm &&
              ${BCMDS[grub]} --efi-directory=${BMNTS[UEFI-grub]} --bootloader-id=$DIST &&
              grub-mkconfig -o /boot/grub/grub.cfg &&
              umount /run/udev && umount /run/lvm"
    fi

    return 0
}

setup_syslinux()
{
    EDIT_FILES[9]="/boot/syslinux/syslinux.cfg"
}

setup_systemd-boot()
{
    EDIT_FILES[9]="/boot/loader/entries/$DIST.conf"
}

setup_boot_device()
{
    infobox "$_PrepMount" "\nSetting device flags for: $BOOT_PART\n" 1

    if [[ $BOOT_PART = /dev/nvme* ]]; then
        BOOT_DEVICE="${BOOT_PART%p[1-9]}"
    else
        BOOT_DEVICE="${BOOT_PART%[1-9]}"
    fi

    BOOT_PART_NUM="${BOOT_PART: -1}"

    if [[ $SYS == 'UEFI' ]]; then
        parted -s $BOOT_DEVICE set $BOOT_PART_NUM esp on >/dev/null 2>&1
    else
        parted -s $BOOT_DEVICE set $BOOT_PART_NUM boot on >/dev/null 2>&1
    fi

    return 0
}

prerun_grub()
{
    local cfg="$MNT/etc/default/grub"
    sed -i "s/GRUB_DISTRIBUTOR=.*/GRUB_DISTRIBUTOR=\"${DIST}\"/g;
            s/GRUB_CMDLINE_LINUX_DEFAULT=.*/GRUB_CMDLINE_LINUX_DEFAULT=\"\"/g" $cfg

    if [[ $LUKS_DEV ]]; then
        sed -i "s~#GRUB_ENABLE_CRYPTODISK~GRUB_ENABLE_CRYPTODISK~g;
                s~GRUB_CMDLINE_LINUX=.*~GRUB_CMDLINE_LINUX=\"${LUKS_DEV}\"~g" $cfg
    fi

    if [[ $SYS == 'BIOS' && $LVM && $SEPERATE_BOOT == false ]]; then
        sed -i "s/GRUB_PRELOAD_MODULES=.*/GRUB_PRELOAD_MODULES=\"lvm\"/g" $cfg
    fi

    # needed for os-prober module to work properly in the chroot
    mkdir -p /run/lvm
    mkdir -p /run/udev
    mkdir -p $MNT/hostrun/lvm
    mkdir -p $MNT/hostrun/udev
    mount --bind /run/lvm $MNT/hostrun/lvm
    mount --bind /run/udev $MNT/hostrun/udev

    return 0
}

prerun_systemd-boot()
{
    # no LVM then systemd-boot uses PARTUUID
    [[ $ROOT_PART =~ /dev/mapper ]] || ROOT_PART_ID="PART$ROOT_PART_ID"

    # create the boot entry configs
    mkdir -p ${MNT}${BMNTS[$SYS-systemd-boot]}/loader/entries
    cat > ${MNT}${BMNTS[$SYS-systemd-boot]}/loader/loader.conf << EOF
default  $DIST
timeout  5
editor   no
EOF
    cat > ${MNT}${BMNTS[$SYS-systemd-boot]}/loader/entries/${DIST}.conf << EOF
title   $DIST Linux
linux   /vmlinuz-${KERNEL}$([[ $UCODE ]] && printf "\ninitrd  %s" "/${UCODE}.img")
initrd  /initramfs-$KERNEL.img
options root=$ROOT_PART_ID $([[ $LUKS_DEV ]] && printf "%s " "$LUKS_DEV")rw
EOF
    # add pacman hook to update the bootloader when systemd receives an update
    mkdir -p $MNT/etc/pacman.d/hooks
    cat > $MNT/etc/pacman.d/hooks/systemd-boot.hook << EOF
[Trigger]
Type = Package
Operation = Upgrade
Target = systemd

[Action]
Description = Updating systemd-boot
When = PostTransaction
Exec = /usr/bin/bootctl update
EOF
    # systemd-boot requires this before running bootctl
    systemd-machine-id-setup --root="$MNT" >/dev/null 2>&1
    return 0
}

prerun_syslinux()
{
    mkdir -pv $MNT${BMNTS[$SYS-syslinux]}/syslinux
    cp -rfv /usr/lib/syslinux/bios/* $MNT${BMNTS[$SYS-syslinux]}/syslinux/

    cat > $cfgdir/syslinux.cfg << EOF
UI menu.c32
PROMPT 0

MENU TITLE $DIST Syslinux Boot Menu
TIMEOUT 50
DEFAULT $DIST

LABEL $DIST
MENU LABEL $DIST Linux
LINUX ../vmlinuz-$KERNEL
APPEND root=$ROOT_PART_ID $([[ $LUKS_DEV ]] && printf "%s " "$LUKS_DEV")rw
INITRD ../initramfs-$KERNEL.img$([[ $UCODE ]] && printf "\nINITRD %s" "../${UCODE}.img")

LABEL ${DIST}fallback
MENU LABEL $DIST Linux Fallback
LINUX ../vmlinuz-$KERNEL
APPEND root=$ROOT_PART_ID $([[ $LUKS_DEV ]] && printf "%s " "$LUKS_DEV")rw
INITRD ../initramfs-$KERNEL-fallback.img$([[ $UCODE ]] && printf "\nINITRD %s" "../${UCODE}.img")
EOF
    return 0
}

install_bootloader()
{
    if ! [[ $ROOT_PART =~ /dev/mapper ]]; then
        ROOT_PART_ID="UUID=$(blkid -s PARTUUID -o value $ROOT_PART)"
    else
        # for LVM we just use the partition label
        ROOT_PART_ID="$ROOT_PART"
    fi

    # for virtual machine remove the ucode
    [[ $VM ]] && UCODE=""

    # remove old UEFI boot entries
    if [[ $SYS == 'UEFI' ]]; then
        find ${MNT}${BMNTS[UEFI-$BOOTLDR]}/EFI/ \
            -maxdepth 1 -mindepth 1 -name '[aA][rR][cC][hH][lL]abs' \
            -type d -exec rm -rf '{}' \; >/dev/null 2>&1
    fi

    prerun_$BOOTLDR
    printf "\nInstalling and setting up $BOOTLDR in ${BMNTS[$SYS-$BOOTLDR]}\n\n"
    chrun "${BCMDS[$BOOTLDR]}" #2>$ERR
    echeck "${BCMDS[$BOOTLDR]}"

    if [[ -d $MNT/hostrun ]]; then
        printf "\n"
        umount $MNT/hostrun/udev >/dev/null 2>&1
        umount $MNT/hostrun/lvm >/dev/null 2>&1
        rm -rf $MNT/hostrun >/dev/null 2>&1
    fi

    # copy efi stub to generic catch all
    if [[ $SYS == 'UEFI' && $BOOTLDR == 'grub' ]]; then
        uefi_boot_fallback
    fi

    return 0
}

uefi_boot_fallback()
{
    # some UEFI firmware requires a dir in the ESP with a generic bootx64.efi
    # see:  https://wiki.archlinux.org/index.php/GRUB#UEFI

    local esp="${MNT}${BMNTS[$SYS-$BOOTLDR]}"

    local default
    default="$(find $esp/EFI/ -maxdepth 1 -mindepth 1 -name '[Bb][oO][oO][tT]' -type d)"

    if [[ $default ]]; then
        default="${default##/*}"
    else
        default="boot"
    fi

    printf "\n"

    if [[ -d $esp/EFI/$default ]]; then
        rm -rfv $esp/EFI/$default/*
    else
        mkdir -pv $esp/EFI/$default
    fi

    cp -fv $esp/EFI/$DIST/grubx64.efi $esp/EFI/$default/bootx64.efi

    sleep 2
    return 0
}

shim_secure_boot()
{
    efibootmgr -c -w -L $DIST -d $BOOT_DEVICE -p $BOOT_PART_NUM -l ${MNT}${BMNTS[$SYS-$BOOTLDR]}/shim64.efi
}
